import sys
import os
from multiprocessing import Pool, cpu_count, Process
import json

def _except_hook(cls, exception, traceback):
    '''
    Function used by :py:func:`pqt5_exception_workaround`. Don't call this directly.
    '''
    sys.__excepthook__(cls, exception, traceback)
    

    
def pqt5_exception_workaround():
    '''
    Workaround to prevent pyqt5 from silently eating exceptions.
    Call this at the beginning of your progame and you will experience relieve.
    '''
    sys.excepthook = _except_hook


class Persival():
    '''
    Keeps values persistent between program runs. 
    
    '''
    
    def __init__(self, persival_file='persival.json'):
        '''
        Will try to load parameters from the persival_file. 
        If the file does not exists this will be silently ignored and the file will be created when :py:meth:`save` will be called. 
    
        '''

        self.persival_file=persival_file        
        if os.path.exists(persival_file):
            with open(self.persival_file, 'r') as file:
                self.data=json.load(file)
        else:
            self.data={}
        
    def setDefault(self, name, value):
        '''
        Sets the given value only if no information was present in the persival file or if the persival file was not found.
        '''
        if not name in self.data.keys():
            self.data[name]=value
    
    def get(self, name):
        '''
        Returns the saved value for the given parameter.
        '''
        return self.data[name]
    
    
    def set(self, name, value):
        '''
        Store the given value in memory. You have to call save to dump the value to the disk. 
        '''
        self.data[name] = value
       # print('Break')
    
        
    def save(self):
        '''
        Saves the content to a file.
        '''
        with open(self.persival_file, 'w') as file: 
            json.dump(self.data, file)
    
    
    
    
class ProcessingHelper():
    '''
    Helper working around the limitations of Pool.
    
    '''
    def __init__(self, parallel=False):
        self.funcs=[]
        self.parallel=parallel
    
    def add(self, function, args, kwargs={}):
        '''
        Add a function to the processing pool.
        
        :param function: function to be executed
        :param args: list. Provice the positional arguments
        :param kwargs: dict. Optional, provide the kwargs
        '''
        self.funcs.append({'func': function, 'args': args, 'kwargs': kwargs})
        
    def clear(self):
        self.funcs=[]
    
    def run(self):
        '''
        Start processing. 
        
        :returns: list of results. The orgdering matches the sequence of add calls.
        '''
        processes=cpu_count()
        # Windows does not handle full cpu loads well, so leave one CPU unutilized for Windows.
        if os.name == 'nt':
            processes -=1
            
        if self.parallel:
            with Pool(processes=processes) as pool:
                result=pool.map(self._myrun, range(len(self.funcs)))
        else:
            result=[]
            for i, f in enumerate(self.funcs):
                r=self._myrun(i)
                result.append(r)
                
        return result
    
    
    def _myrun(self, paramNo):
        '''
        wrapper around the function for pool
        
        :param paramNo:
        '''
        param=self.funcs[paramNo]
        func=param['func']
        args= param['args']
        kwargs= param['kwargs']
        retval=func(*args, **kwargs)
        return retval
        


'''
    File name: fileReader.py
    Author: Simon Waid
    Date created: 2021
    Python Version: 3.8, 3.9
'''

import xml.etree.ElementTree as ET

import os
import logging
import numpy as np
import sys
import time
import math
from numba import njit
#import vaex
import hashlib
import numpy as np
import matplotlib.pylab as plt
import scipy.integrate as integrate
import json
import pickle

#from tool_box import ProcessingHelper
import pandas as pd

class Osci_XML_Tags():
    '''This is an empty class for making the oscilloscope configuration stored in the xml accessible'''
    def __init__(self):
        pass
    
class MyCache:
    '''
    Implements caching for :py:class:`Det_File`
    '''
    ENABLE_IN_RAM = False
    def __init__(self, cache_dir, instance_id):
        '''
        :param class_hash: Data that make the instance unique.
        '''
        self.cache_dir = cache_dir
        if not os.path.exists(cache_dir):
            os.mkdir(self.cache_dir)
        self.instance_id=instance_id.encode('utf-8')
        self.data={}

    def cache(self, func):
        '''
        Decorator for cachings.
        ''' 
        def cache_func(*args, **kwargs):
            hash = hashlib.sha1()
            # Include data about then instance 
            hash.update(self.instance_id)
            # Skip self.
            # TODO: Improve this.
            args = list(args)
            for i in range(len(args)):
                if type(args[i]) == pd.DataFrame:
                    args[i] = args[i].to_json()
            cache_key = json.dumps([func.__name__, args, sorted(kwargs)], separators = (',', ':'))
            hash.update(cache_key.encode('utf-8'))
            my_hash = hash.hexdigest()
            # Access the in-RAM cache first
            if hash in self.data.keys():
                
                print(f'In-RAM-Cache hit {func.__name__}, hash: {my_hash}')
                return self.data[my_hash]
            else:
                hashFileName = os.path.join(self.cache_dir, my_hash + '.pickle')
                # Try to access the on Disk cache.
                if os.path.exists(hashFileName):
                    try:
                        with open(hashFileName, 'rb') as f:
                            result=pickle.load(f)
                        print(f'On-Disk-Cache hit {func.__name__}, hash: {my_hash}')
                    except:
                        print(f'Failed to load from disk-cache. Consider implementing an automated handling of this situation.,')
                        raise
                else:
                    print(f'Cache miss {func.__name__}, hash: {my_hash}')
                    # Execute the function an store the data in the on-disk cache
                    t = time.time()
                    result = func(*args, **kwargs)
                    
                    print('Computation took {:.2}s'.format(time.time() - t))
                    with open(hashFileName, 'wb') as f:
                        pickle.dump(result, f)
                    print(f'Stored in cache {my_hash}')
                # Store in the In-RAM cache.
                if self.ENABLE_IN_RAM:
                    self.data[hash] = result
                
                return result
            
        return cache_func
                
class RS_Analysis():
    '''
    Extends RS_File with analysis functions for detectors.
    '''
    BUFFER = 5e-8
    STD_FACTOR=3
    def __init__(self, rs_file, cache_dir = None):
        '''
        :param rs_file: Instance of :py:class:`RS_file`
        :param cache_dir: Optional. If given, some computationally expensive functions will use a cache. This requires joblib. 
        '''
        self.rs_file=rs_file
        # Initialize the cache
        # Change this rev. number to invalidate the cache
        rev = 1
        if not cache_dir is None:
            instance_id = str(self.rs_file.meta)+str(rev)
            self.memory = MyCache(cache_dir, instance_id)
            # Overwrite the functions for which we would like to use the cache.
            self.__getTot = self.getTot
            self.getTot = self.memory.cache(self.__getTot)
            
            self.__getStatistics = self.getStatistics
            self.getStatistics=self.memory.cache(self.__getStatistics)
            
            self.__fastStat = self.fastStat
            self.fastStat=self.memory.cache(self.__fastStat)
        
        else:
            self.memory = None

    def fastStat(self, sources=None):
        '''
        Fast calculation of statistics over the full sample.
        :TODO: Implement multiprocessing.
        
        :param sources:
        
        :returns: dict containing statistics.
        '''
        
        if sources is None:
            sources= self.rs_file.meta['source_names']

        result={}
        # Calculation loop.
        # Should be parallelized
        std_res=[]
        for source in sources:
            raw_data=self.rs_file.data_raw_RS[source]
            # Depending on the data format use a different method for calculating the statistics.
            # If we have raw adc counts we use  _stat_raw.
            # If we have floats we use  _stat.
            if self.rs_file.meta['apply_offset']:
                offset =self.rs_file.meta['offset'][source]
                scaling = self.rs_file.meta['conversion_factor'][source]
                # Calculate the standard deviations
                avg=np.average(raw_data)
                offset_signal=avg
                data_ch = _stat_raw(raw_data, offset, offset_signal, scaling)
                std_res.append(data_ch)
 
                #  p.add(_one_side_std_raw, [raw_data, offset, scaling])
            else:
                offset=np.average(raw_data)
                std_res.append(_stat(raw_data, offset))
                #    p.add(_one_side_std, [raw_data])
        
        # Evaluation loop
        for i, source in enumerate(sources):
            
            stdA, stdB, my_min, my_max, offset= std_res[i]
            # Out standard deviation for the threshold is the minimum. 
            std=min([stdA, stdB])
            result[source]={}
            result[source]['std'] = std
            result[source]['stdA'] = stdA
            result[source]['stdB'] = stdB
            result[source]['min'] = my_min
            result[source]['max'] = my_max
            result[source]['offset'] = offset
        
        return result
 
    def getStatistics(self, acquisition=None, start = None, stop = None):#, source=None):
        '''
        Returns some statistics on the data. This is a slow but reliable implementation. Consider using :py:meth:`.fastStat`
        
        :param start: See :py:meth:`.getAsDf` 
        :param stop: See :py:meth:`.getAsDf`
        '''
        #:param source: Oscilloscope Source Name
        
        #raise(RuntimeError('This function is buggy, please fix is p'))
        
        # Get filtered data
        data = self.rs_file.getRaw(acquisition=acquisition, start = start, stop = stop, source=None)
        
        result = {}

        # We operate on one channel at a time to keep the memory footprint low.
        for i, ch in enumerate(self.rs_file.meta['source_names']):

            result[ch] = {} 
            
            # Convert the data to floating point.
            # We generate a dict that looks like the raw_data dict so we can use the same functions.
            data_ch={ch:data[ch]} 
            data_ch = self.rs_file.rawToDtypeOut(data_ch)[ch]

            # Check over the whole sample whether peaks are on the negative or the positive 
            # side. 
            my_min = np.amin(data_ch)#.astype(np.int64)
            my_max = np.amax(data_ch)#.astype(np.int64)
            
            result[ch]['min'] = my_min
            result[ch]['max'] = my_max
                        
            # If only fast tasks were requested we're done!
            # Otherwise it's time to calculate the standard deviation.
            # Fast am memory efficient implementation of std calculation.
            # Apply a filtering only using values on the other side of the peaks to calculate the std
            if np.abs(my_min) > np.abs(my_max):
                mask= data_ch<=0
            else:
                mask= data_ch>=0
            
            masked=data_ch[mask]
            del mask
            # Time and memory efficient rms over the masked entries 
            masked=np.square(masked)
            std = np.sqrt(np.mean(masked))
            del masked
            
            result[ch]['std'] = std
            # Convert output 
            
        return result
    
    def get_2d_histo(self, source, start = None, stop = None, no_bins_x=8000, max_bins_y=1024, cutoff_y=1E-5):
        '''
        Returns a 2D histogram of the data. This is useful for plotting acquisitions with a large number of samples.
        
        :param source: Oscilloscope Source Name. 
        :param start: Index of starting sample
        :param stop: Index of final sample
        :param no_bins_x: Number of bins in x direction. Optional, defaults to 8192 (fits an 8k screen, the largest type of screen currently on the market).
        :param max_bins_y: Maximum number of bins in y direction. Optional, defaults to 1024. The resolution of the ADC will be inferred from the data. The number of y bins will be adapted to the inferred resolution.   
        :param cutoff_y:
        
        :returns: histogram, extent where histogram is a 2d numpy array and extent contains the limits in physical units for use with matplotlib imshow.
                
        '''
        
        data = self.rs_file.getRaw(start = start, stop = stop, source=source)
        data=data[source]
        # We need to parameterize our 2D output array first.
        # We start with the number of y bins. 
        # We compute the unique values first.
        unique=np.unique(data)
        # And find find the minimum distance between unique values.
        dist_y=np.min(unique[1:]-unique[:-1]).astype(np.float64)
        # Assuming the ADC is linear we now know it's resolution.
        # Now let's get the minimum and maximum values so we known the swing of the data.
        max_val=np.max(data).astype(np.float64)
        min_val=np.min(data).astype(np.float64)
        
        print(min_val, max_val)

        # We now can compute the number of needed bins.
        span=max_val-min_val
        no_bins_y=int(span/dist_y)
        # We might have more bins than the ADC has levels. In that case restrict to the number of ADC levels
        if  no_bins_y > self.rs_file.meta['quantisation_levels']:
            no_bins_y = self.rs_file.meta['quantisation_levels']
        
        # If we need to restrict ourselves, we first attempt to remove outliers.
        if no_bins_y > max_bins_y:
            # To find outliers we create a 1D histogram 
            histo_1d_occ, histo_1d_val=np.histogram(data, bins=no_bins_y)
            # We find the maximum occurrence 
            histo_1d_val=histo_1d_val[:-1]
            max_occ=np.max(histo_1d_occ)
            # Compute the threshold and eliminate everything below the threshold
            threshold=max_occ*cutoff_y
            
            above_th_bool=histo_1d_occ > threshold
            #where_above_th=np.where(above_th_bool)
            above_th=histo_1d_val[above_th_bool]
            # Update the limits of the y part of interest
            max_val=np.max(above_th).astype(np.float64)
            min_val=np.min(above_th).astype(np.float64)
            span=max_val-min_val
            no_bins_y=int(span/dist_y)
            print(threshold, min_val, max_val)
            
        if no_bins_y > max_bins_y:
            no_bins_y = max_bins_y
            
        
        # Update the distance between samples
        dist_y=span/no_bins_y
        
        # In x direction the number of bins is simply limited by the number of samples:
        if len(data) < no_bins_x:
            no_bins_x=len(data) 
        
        #if 'X-Time' in self.rs_file.meta['channel_type']:
            #raise(RuntimeError('Not implemented for explicit time scales'))
        #    time = self.rs_file.getRaw(start = start, stop = stop, source='X-Time')
        #    histogram=np.histogram2d(data, time, bins=[no_bins_x, no_bins_y])
        #else:
        histogram=_histo_implicit_t(data, no_bins_x, no_bins_y, min_val, dist_y)
        # Now create the histogram
        
        if self.rs_file.meta['apply_offset']:
            y_min=min_val*self.rs_file.meta['conversion_factor'][source]+self.rs_file.meta['offset'][source]
            y_max=max_val*self.rs_file.meta['conversion_factor'][source]+self.rs_file.meta['offset'][source]
        else:
            y_min=min_val
            y_max=max_val
    
        #x_axis=np.linspace(self.meta['xStart'], self.meta['xStop'], no_bins_x)
        #y_axis=np.linspace(y_min, y_max, no_bins_y)
        
        if not start is None or not stop is None:
            raise(RuntimeError('Not implemented!'))
             
        extent=[self.rs_file.meta['xStart'], self.rs_file.meta['xStop'], y_min, y_max]
            
        return histogram, extent # x_axis, y_axis
        
    def plotEvents(self, events, dir_out, sources= None, no_events = 10, buffer = BUFFER):
        '''
        :param events: Output form getTot
        :param sources: The oscilloscope Source that should be processed.
        :param dir_out: Directory for plotting.
        :param no_events: Maximum number of events that should be plotted.
        :param buffer: How much data should be plotted before and after the event. Defaults to ns.
        '''
        # If the directory does not exist create it
        if not os.path.exists(dir_out):
            os.mkdir(dir_out)
        # 
        if sources is None:
            sources= self.rs_file.meta['source_names']
        # If we didn't get a list, assume it's a single source and create a list 
        elif not type(sources) == type([]):
            sources =[sources]
            
        for i, source in enumerate(sources):
            
            ev=events[source]
            # If no events were detected silently skip plotting
            if ev is None:
                continue
            # make local variable for all relevant data.
            start = ev['Start'].values
            length = ev['Duration'].values
            t_sample = self.rs_file.meta['t_sample']
            source_names = source
            buffer_raw = int(buffer / t_sample)

            plot_color = ['blue', 'green', 'orange', 'black']
    
            # Limit the plotting to the first events as specified by no_events
            if len(start) > no_events:
                start = start[:no_events]
                length = length[:no_events]
                
            peak_count = 1
            for s, l in zip(start, length):
                start = s - buffer_raw
                stop = s + l + buffer_raw
                if start < 0:
                    start = 0
                if stop > (self.rs_file.total_no_samples):
                    stop = int(self.rs_file.total_no_samples)
                try:
                    data = self.rs_file.getAsDf(start=start, stop=stop, time=True)
                except:
                    print('Breakpoint!')
                    raise
                time = data['Time']
                # change axis to begin at position of the first sample exceeding the threshold
                time = time - s*t_sample
                voltage = data[source_names]
    
                plt.close()
                plt.clf()
                plt.plot(time, voltage, '-', color = plot_color[i])
    
                plt.grid()
                #plt.legend(title='Parameter where:')
                plt.title(f'{str(source)} - Peak_no: {str(peak_count)}')
                outfile = os.path.join(dir_out, ('peak' + str(peak_count)+'.png'))
                plt.savefig(outfile, dpi = 300)
                peak_count += 1
                plt.close()

    def get_threshold(self, stats):
        '''
        Returns the threshold for each channel based on a given dict including max, min and std values as 
        we get it for each channel from _getStatistics
        :param stats: Output from getStatistics including max, min and std
        '''
        std_factor = 6
        result = {}
            
        for source, values in stats.items():
            threshold = 0
            if abs(values['min']) > abs(values['max']):
                threshold = -std_factor * values['std']
            else:
                threshold = std_factor * values['std']
            result[source] = threshold

        return result

    def getTot(self, source, threshold, cmp = None, signal_samples = None, tmin_gap = BUFFER):
        '''

        Get the time over threshold for signals. Analyses the peak signals over the given
        threshold and determines where peaks occured and how long they lasted. Returns a dataframe 
        containing the start indices and the length (in samples) reffering to the raw data. It includes
        a fine tuning process, changing the minimal required signal length time according to the found
        data. 

        :param source: The source name on which the operation should be carried out.
        :param threshold: The threshold for detecting an event. Unit: Volts.
        :param cmp: Either '>' or '<'. Determines if values above or below the threshold should be considered
        :param gap_time: Search limit for the minimal required time between two signals. Unit: seconds
        :param signal_time: Initial search limit for the minimal required time to be considered as signal. Unit: seconds
        
        :returns: Pandas DataFrame. The Columns are Start and Duration. They contain the beginning and end of peaks in sample indices.  
        '''
        
        if cmp is None:
            if threshold < 0:
                cmp = '<'
            else:
                cmp = '>'

        t_sample = self.rs_file.meta['t_sample']
        
        # Set the minimal signal time to at least three samples, if not given specifically
        if signal_samples == None:
            tmin_signal = 2 * t_sample
        else:
            tmin_signal = signal_samples * t_sample

        data = self.rs_file.data_raw_RS[source]
        
        if np.isnan(threshold):
            raise(RuntimeError('Invalid threshold. NaN is now allowed'))
        
        # Convert the threshold to raw data units. 
        if self.rs_file.meta['apply_offset']:
            threshold_raw = threshold - self.rs_file.meta['offset'][source]
            threshold_raw = threshold_raw / self.rs_file.meta['conversion_factor'][source]  
            threshold_raw = np.asarray(threshold_raw, dtype=data.dtype)
        else:
            threshold_raw = threshold

        if tmin_signal < t_sample:
            raise(RuntimeError('Initial signal length is smaller than the sample size!!!'))

        if cmp == '>':
            if threshold < 0:
                logging.warn(f'threshold is below 0 but cmp is >. This is an unexpected combination.')
            det = data >= threshold_raw
        elif cmp == '<':
            if threshold > 0:
                logging.warn(f'threshold is below 0 but cmp is >. This is an unexpected combination.')
            det = data < threshold_raw
        else:
            raise(RuntimeError('Invalid value for cmp'))
            
        # Free memory
        del data 

        det = det.astype(np.int8)
        # Theory of operation: 
        # Lets assume we have the following in a:
        # a = [0 0 0 1 1 1 1 0 0]
        # We shift by 1 to the left:
        # b = a[1:]
        # b = [0 0 1 1 1 1 0 0]
        # Now let's subtract:
        # a[:-1] - b = [ 0 0 -1 0 0 0 1 0]
        
        det2 = det[:-1] - det[1:]
        # Free memory
        del det

        # Get indices of starting and ending points for each peak
        starts_in = np.where(det2 == -1)
        ends_in = np.where(det2 == 1)
        starts = starts_in[0]
        ends = ends_in[0]
        
        # Ensure the start and ends to be in sync
        if len(starts) < 1:
            return None
        if starts[0] > ends[0]:
            ends = ends[1:]
        if len(starts) > len(ends):
            starts = starts[:len(ends)]

        # Free memory
        del det2

        # Find the corresponding peaks according to some required settings 
        #starts_out, ends_out = _peak_filter(starts, ends, tmin_signal_raw, tmin_gap_raw)
        starts_out=starts
        ends_out=ends

        # Convert to numpy
        starts_out = np.array(starts_out)
        ends_out = np.array(ends_out)
        
        # Now convert starts and ends to starts and length of signals
        result = pd.DataFrame(starts_out, columns=['Start'])
        try:
            result['Duration'] = ends_out - starts_out
        except:
            print('Len start: {}, Len end: {}'.format(len(starts_out), len(ends_out)))
            raise
        result['Duration_SI'] = result['Duration']*self.rs_file.meta['t_sample']
        if np.any(result['Duration'] <= 0):
            raise(RuntimeError("Negative durations don't exist. This is a bug."))

        return result

    def analyse_peaks(self, events, source, cmp, buffer = BUFFER):
        '''
        Analyze single peaks. Takes an events dataframe as returned by the function getTot_fine()
        and returns a dataframe containing the maximal peak heigthh and the peak areas of each event. Has 
        to be launched for each file and channel.
        The method covers different signs of the voltage, such that returned values are always positive 
        and refer to the absolute values of height and area.
    
        :param events: Starting indices and length in samples of the found events from getTot_fine().
        :param source: The Souce  number on which the operation should be carried out.
        :param cmp: < means too look for maxima in the negative direction, > otherwhise.
        :param buffer: Event locations come at time where threshold is exceeded and underpassed again. 
        The actual peak takes longer than that, therefore we add a buffer on both ends. 
        '''  
        ch = self._chNameToNum(source)
        t_sample = self.rs_file.rs_file.meta['t_sample']
        buffer_raw = int(buffer / t_sample)

        results = pd.DataFrame(columns = np.array(['max_height', 'area']))
        max_height = []
        area = []

        if len(events) == 0:
            return None

        start = events['Start'].values
        length = events['Duration'].values
        
        dataList = []
        for s, l in zip(start, length):
            start = s - buffer_raw
            stop = s + l + buffer_raw
            if start < 0:
                start = 0
            if stop > (self.rs_file.meta['length'] - buffer_raw):
                stop = int(self.rs_file.meta['length'] - 1)
            length = stop - start
            timestamps = np.linspace(start, stop, length, dtype = np.float64)
            data = self._getRaw(start, stop, sourceNo = ch)
            data = self._rawToDtypeOut(data, sourceNo = ch)
            
            dataList.append([timestamps, data])

        for data in dataList:
            mytime = data[0]
            voltage = data[1]

            # Check for the direction of the signal voltage and adapt accordingly 
            # The integration method can of course be changed. The simple trapezoidal method is used
            # currently. However, one can change this to simpson() or even higher order integrations.
            if cmp == '<':
                max_height.append(-min(voltage))
                area.append((-integrate.trapz(y = voltage, x = mytime)) * t_sample)
            else :
                max_height.append(max(voltage))
                area.append((integrate.trapz(y = voltage, x = mytime)) * t_sample)
        
        results['max_height'] = np.array(max_height)
        results['area'] = np.array(area)

        return results

    def getTotAuto(self, sources=None):
        '''
        Convenience function to get the tot with reasonable assumptions. Calls :py:meth:`.getTot_fine`
        
        :param sources: List of sources to be processed. 
        :returns: results, meta. results is a dict containing the results. the source name is the key. meta is a dict containing meta data such as the standard deviations. 
        '''
        
        # TODO: Implement offset detection and elimination  
        
        if sources is None:
            sources= self.rs_file.meta['source_names']

        # Get the statistics.
        t=time.time()
        
        #stats = self.getStatistics(source=sources)
        stats = self.fastStat(sources)
        print('Statistics took: ', time.time()-t)
        
        # Calculate the time over threshold
        meta=stats
        for source in sources:
            # Get the result from the previous (parallel) calculation
            std=stats[source]['std']
            stdA=stats[source]['stdA']
            stdB=stats[source]['stdB']
            my_min=stats[source]['min']
            my_max=stats[source]['max']
            # Get the threshold.
            threshold= std*self.STD_FACTOR
            # Adjust the sign of threshold.
            # Using the min and max value to determine the sign seems best.
            # An alternative would be to compare the two standard deviations.
            # However, the latter method can be inaccurate due to numeric issues.
            # We start by comparing min and max.  If there is sufficient difference, we use min and max.
            # If not, we use the stds
            ma= max([abs(my_min), abs(my_max)])
            mi= min([abs(my_min), abs(my_max)])
            if abs(ma/mi-1) <1E-3:
                # Difference between min and max too small 
                if stdA > stdB:
                    threshold*=-1
            elif abs(my_min) > abs(my_max):
                
                threshold*=-1
            print(f'Source: {source} std {std}, min {my_min}, max {my_max} ==> threshold: {threshold}')
            meta[source]['threshold']=threshold
            
        # Calculate the time over threshold 
        results={}

        for source in sources:
            
            results[source] = self.getTot(source, threshold)
            
        return results, meta 

@njit
def _histo_implicit_t(data, no_bins_x, no_bins_y, min_y, dist_y):
    '''
    Compute a histogram of the given data. It assumes there are no time stamps available in the data. If time stamps are available, the numpy histogram2d function can be used.
    
    '''

    # Initialize the histogram. For now  we have zero counts every
    output=np.zeros((no_bins_x, no_bins_y))
    
    # Determine the distance between bins in the x direction.  
    length=len(data)
    dist_x=length/no_bins_x

    #Multiplications are cheaper than divisions, so divide once. Later on use the inverse 
    dist_y_inv=1/dist_y
    
    # Initialize variables for the loop. 
    x_index=0
    x_index_limit=dist_x
    for index, d in enumerate(data):
        # Determine the location of our sample in x direction
        if index > x_index_limit:
            x_index+=1
            x_index_limit= (x_index+1)*dist_x
        # Determine the location of our sample in x direction
            
        y_index=int(float(d-min_y)*dist_y_inv)
        # Avoid out of range indexing
        if y_index  < 0  or y_index > no_bins_y:
            continue
        # Update the histogram
        output[x_index, y_index] +=1
    
    return output
    
@njit
def _stat_raw(data, offset, offset_signal, scaling):
    '''
    Compute statistics from raw data.

    '''
    
    sideA_sum=float(0)
    sideB_sum=float(0)
    sideA_len=float(0)
    sideB_len=float(0)
    
    l=len(data)
    
    my_min=float(data[0]-offset_signal)*scaling+offset
    my_max=float(data[0]-offset_signal)*scaling+offset
    
    for d in data:
        #df=float(d)
        df=float(d-offset_signal)*scaling+offset
        if df <= 0:
            sideA_sum +=df**2
            sideA_len +=1
        if df >= 0:
            sideB_sum +=df**2
            sideB_len +=1
        if df < my_min:
            my_min=df
        if df > my_max:
            my_max=df
             
    if sideA_len == 0:
        stdA = np.nan
    else:     
        stdA = math.sqrt(sideA_sum/sideA_len)
    if sideB_len == 0:
        stdB = np.nan
    else:
        stdB = math.sqrt(sideB_sum/sideB_len)
    
    offset_signal_ret=float(offset_signal)*scaling+offset
        
    return stdA, stdB, my_min, my_max, offset_signal_ret
    
@njit
def _stat(data, offset_signal):
    '''
    Compute statistics from floating point numbers     
    
    '''
    
    sideA_sum=float(0)
    sideB_sum=float(0)
    sideA_len=float(0)
    sideB_len=float(0)
    
    my_max=data[0]
    my_min=data[0]
    
    for d in data:
        d-= offset_signal
        if d <= 0:
            sideA_sum +=d**2
            sideA_len +=1
        if d >= 0:
            sideB_sum +=d**2
            sideB_len +=1
        if d < my_min:
            my_min=d
        if d > my_max:
            my_max=d
    
    # Calculate std. Avoid division by 0 
    if sideA_len == 0:
        stdA = np.nan
    else:
        stdA = math.sqrt(sideA_sum/sideA_len)
    if sideB_len == 0:
        stdB = np.nan
    else:
        stdB = math.sqrt(sideB_sum/sideB_len)
    
    return stdA, stdB, my_min, my_max, offset_signal
    
@njit
def _peak_filter(starts, ends, lenght_min, tmin_gap_raw):
    '''
    Filters the detected peaks. For use by :py:meth:RS_Det_File.getTot_fine.
        
    :param starts: list of indices with a potential start of a peak
    :param ends: list of indices with a potential end of a peak. Must match starts
    :param lenght_min: minimum number of samples for a peak to valid
    :param tmin_gap_raw: Minimum gap beteen peeks.
        
    :returns starts_out, ends_out: Filtered lists like starts and ends
    '''
    starts_out = []
    ends_out = []
    
    # Init for the loop.
    max_i = len(starts)
    i = 0
    while(True):
        length = ends[i] - starts[i]
        # First condition: The distance between the start and end must be sufficient. 
        if length >= lenght_min:
            start = starts[i]
            starts_out.append(start)
            # Now find the end.
            while(True):
                end = ends[i]
                # Is the next start within tmin_gap_raw from the current end?
                # If so, continue searching.
                # There is one exception to it: We're already at the end.
                # If we reach the last entry stop.
                if i + 1 >= max_i:
                    ends_out.append(end)
                    break    
                
                if end + tmin_gap_raw > starts[i+1]:
                    i += 1
                else:
                    ends_out.append(end)
                    break
        i += 1
        if i >= max_i:
            break
    
    return starts_out, ends_out        
        
        


'''
    File name: RS_file.py
    Author: Simon Waid
    Date created: 2021
    Python Version: 3.8, 3.9
'''


import xml.etree.ElementTree as ET
import os
import logging
import numpy as np
import pandas as pd
import numpy as np
import binascii

class RS_File():
    '''
    Reads Rhode & Schwarz RTP16 files. Currently limited to .csv and .bin files. The oscilloscope writes two files. 
    In case of bin files this is a *.bin and a *.Wfm.bin file. The first once contains metadata. 
    The second the measurement data. Both files are needed. 
    They must be in the same directory and follow the R&S naming convention.
        
    The file format is described here: https://www.rohde-schwarz.com/webhelp/RTP_HTML_UserManual_en/Content/42894a16e1e94a45.htm
       
    Currently only normal waveform data is supported. Other data will be rejected. 
        
    :property xml_tags: Dictionary. The keys are the 'Name' attributes and the 'Values' attributes from the .xml file. This is a convenient way of accessing the most interesting part of the xml file. 
    :property xml_root: The metadata from the .xml file. This an instance of .. .
    :property meta: Dictionary containing meta data. 
    
    '''
    # We work with a lot of data. Theoretically using vaex would be a good idea.
    # Practically vaex does not work with 32 bit floats and plotting with PyQt and vaex is a pain.
    # So for the moment vaex is an option that is disabled. 
        
    USE_VAEX = False
    if USE_VAEX: 
        DTYPE_OUT = np.float64
    else:
        DTYPE_OUT = np.float32
    
    def __init__(self, file_name):
        '''
        :param file_name: The file name without extension.
        '''
        # Read the metadata first. We provide the data to the user via the .xml property.
        # If the data is anything else than a normal waveform (TraceType is eRS_TRACE_TYPE_NORMAL) we reject the file.
        
        file_base, file_extension = os.path.splitext(file_name)
        self.metadata_file = file_name
        self.waveform_file = file_base + '.Wfm' + file_extension
        # Now decode the file. 
        # If this takes too long one should consider not executing this during the class initialization.
        if file_extension == '.bin':
            _, self.xml_tags = self._loadXml(self.metadata_file)
            logging.info(f'Loaded xml file {self.metadata_file}.')
            self._data = None
            self.meta = self._decodeBin(self.xml_tags)
        elif file_extension == '.csv':
            logging.info(f'Processing {self.metadata_file}.')
            self.csv_meta = self._loadCsvMeta(self.metadata_file)
            self._data, self.meta=self._decodeCsvWfm(self.waveform_file, self.csv_meta)
        else:
            raise(RuntimeError('File type not implemented.'))

        self.meta['metadata_file'] = self.metadata_file

    def _genTimeAxis(self, acquistion, start, stop):
        '''
        Generate a time axis for the given sample interval
        
        :param start: Start of the interval as sample number
        :param stop: End of the interval as sample number
        '''
        
        # Things get easier if we have proper numbers as start and end 
        if start is None:
            start =0
        if stop is None:
            if acquistion is None:
                stop = self.total_no_samples 
            else:
                stop = self.samples_per_acquisition    
                
        # Calculate start and stop in the time domain
        xStart = self.meta['xStart'] + start*self.meta['t_sample']
        xStop = self.meta['xStart'] + stop*self.meta['t_sample']
            
        length = stop - start
        #t=time.time()
        
        timestamps = np.linspace(xStart, xStop, length, dtype = np.float64)
        return timestamps
    
    def _loadXml(self, xml_file):
        '''
        Load the xml meta data
        '''
        xml_root = ET.parse(xml_file).getroot()
        xml_tags= {}
        # Iterate over all elements in the xml tree and process the attribues.
        # At the end we generate a flat dictionary with all attributes of the xml tree disregarding the hirarchy.
         
        for elem in xml_root.iter():
            if 'Name' in elem.attrib:
                k=elem.attrib['Name']
                if k in xml_tags:
                    # We don't overwrite keys in our flattened metadata structure. 
                    # If duplicate keys would appear we fail.
                    raise(RuntimeError('Key already exists. This is unexpected. Ciao!'))
                else:
                    xml_tags[k] = elem.attrib
        
        return xml_root, xml_tags

    def forgetRaw(self):
        self._data = None

    def _decodeCsvWfm(self, file_name, csv_meta):
        '''
        Decodes csv data. 
        '''
        #noChannels=int(csv_meta['MultiChannelSource'][0])
        col_names=[]
        # It seems we always have to handle single and multiple channels separately. 
        if csv_meta['SignalFormat'] == 'XYDOUBLEFLOAT':
            col_names.append('Time')
        if csv_meta['MultiChannelExport'][0] == 'Off':
            col_names.append(csv_meta['Source'][0])
        elif csv_meta['MultiChannelExport'][0] == 'On':
            for name in csv_meta['MultiChannelSource'][1:]: 
                # It seems we can get a channel listed twice but only once in the data.
                # So if a channel is listed multiple times we ingore it :(.
                if not name == 'None' and not name in col_names:
                    col_names.append(name)
        else:
            raise(RuntimeError('This is a bug!'))
        
        # Load the waveform data
        csv_data=pd.read_csv(file_name, names=col_names, header=None)
        
        meta={}
        
        meta['length_total'] = len(csv_data)
        meta['length_acquisition'] = int(csv_meta['RecordLength'][0])
        meta['xStart'] = float(csv_meta['XStart'][0])
        meta['xStop'] = float(csv_meta['XStop'][0])
        meta['t_sample'] = (meta['xStop']-meta['xStart'])/meta['length_acquisition']
        # This attribute does not exist for csv data 
        meta['leading_samples']=0
        meta['data_format'] = 'pandas'
        # CSV data never needs the application of an offset or scaling factor
        meta['apply_offset'] = False
        
        meta['source_names'] =col_names
        meta['channel_type'] = []
        if csv_meta['SignalFormat'] == 'XYDOUBLEFLOAT':
            meta['channel_type'] = 'X-Time'
        for c in col_names:
            meta['channel_type'].append('Y')
        
        # Match  the data format of binary files.
        csv_data=csv_data.to_dict(orient='series')
        # Get rid of the Series and get numpy arrays instead
        csv_data={k:v.values for (k,v) in csv_data.items() }
        return csv_data, meta 
    
    def _loadCsvMeta(self, file_name):
        '''
        Loads the csv meta data into memory. 
        '''
        metadata = {}
        with open(file_name, 'r') as f:            
            for line in f:
                #line=line.replace('\n', '')
                m=line.split(':')
                if m[0] in metadata:
                    raise(RuntimeError('Bug!'))
                else: 
                    metadata[m[0]] = m[1:-1]
        return metadata
        
    def _decodeBin(self, xml_tags):
        '''
        Decodes binary data and stores the metadata in memory. 
        '''
        # There seem to be many different input formats.
        # In the following we assemble the format for numpy.
        # We're defining the options for decoding the binary data here.
        # Note: There are many parameters which might have an impact.
        # It is rather obscure which ones do and how. Consequently,
        # for the moment we simply assert certain configuration parameters.
        # It is not clear what happens if the 'UserValue' attribute contradicts the 'Value' attribute. 
        # It seems that typically the 'Value' attribute takes precedence.
        
        # Asserting some User values. This might not be needed.
        logging.debug(f'UserValue SignalFormat: {xml_tags["SignalFormat"]}')
        
        logging.debug(f'UserValue ByteOrder: {xml_tags["ByteOrder"]}')

        # Asserting the byte order as we don't have any samples with another byte oder.
        # Perhaps there is a bug in the oscilloscope. The byte order does not seem to change.
        # This might cause the one or the other surprise in the future.
        if xml_tags['ByteOrder']['Value'] == 'eRS_BYTE_ORDER_MSB_FIRST':
            byte_order='<'
        elif xml_tags['ByteOrder']['Value'] == 'eRS_BYTE_ORDER_LSB_FIRST':
            byte_order='<'
        else:
            raise(RuntimeError(f"Invalid ByteOrder {xml_tags['ByteOrder']['Value']}"))
        
        # Handle single/multi channel formats. 
        # We need the number of columns in the data and their names. 
        source_names = []
        if xml_tags["MultiChannelExport"]['Value']=='eRS_ONOFF_OFF':
            no_columns = 1
            source_names.append(xml_tags["Source"]['Value'])
            #channel_names.append(xml_tags["Source"]['Value'])
        elif xml_tags["MultiChannelExport"]['Value']=='eRS_ONOFF_ON':
            channel_names = []
            no_columns = 0
            for ch in ['I_0', 'I_1', 'I_2', 'I_3']:
                if self.xml_tags["MultiChannelExportState"][ch]=='eRS_ONOFF_ON':
                    source_names.append(self.xml_tags["MultiChannelSource"][ch])
                    channel_names.append(ch)
                    no_columns+=1
        else:
            raise(RuntimeError('Unknown value for MultiChannelExport.'))
        
        meta = {}
        meta['length_raw_expected'] = int(xml_tags['SignalHardwareRecordLength']['Value']) 
        meta['leading_samples'] = int(xml_tags['LeadingSettlingSamples']['Value'])

        # Shorten the column names. We don't need any eRS etc.
        source_names = [x.replace('eRS_SIGNAL_SOURCE_','') for x in source_names]
        meta['source_names']=source_names
        # Convert the data format into something numpy can use.
        # Also determine if the offset and position shall be applied
        encoding_format=[]
        meta['channel_type']=[]
        
        # Enable calculation of the expected file size 
        bytes_per_sample=0
        if xml_tags['SignalFormat']['Value'] == 'eRS_SIGNAL_FORMAT_XYDOUBLEFLOAT':
           
            # There seems to be a bug in the oscilloscope. The byte order is wrong for this data type.
            byte_order='<'
            encoding_format.append(('Time', byte_order+'f8'))
            if xml_tags['XAxisTDRDomain']['Value'] == 'eRS_TDR_TDT_RESULT_SIGNAL_DOMAIN_TIME':
                # Store the type of data in the wfm file 
                meta['channel_type'].append('X-Time')
                bytes_per_sample+=8
            else:
                raise(RuntimeError('X-axis domain not implemented'))
            for ch in source_names:
                bytes_per_sample+=4
                encoding_format.append((ch, byte_order+'f4'))
                meta['channel_type'].append('Y')
            meta['apply_offset']=False
        elif xml_tags['SignalFormat']['Value'] == 'eRS_SIGNAL_FORMAT_FLOAT':
            for ch in source_names:
                bytes_per_sample+=4
                encoding_format.append((ch, byte_order+'f4'))
                meta['channel_type'].append('Y')
            meta['apply_offset']=False
        elif xml_tags['SignalFormat']['Value'] == 'eRS_SIGNAL_FORMAT_INT8BIT':
            for ch in source_names:
                bytes_per_sample+=1
                encoding_format.append((ch, byte_order+'i1'))
                meta['channel_type'].append('Y')
            meta['apply_offset'] = True
        elif xml_tags['SignalFormat']['Value']=='eRS_SIGNAL_FORMAT_INT16BIT':
            for ch in source_names:
                bytes_per_sample+=2
                encoding_format.append((ch, byte_order+'i2'))
                meta['channel_type'].append('Y')
                #quantisation_levels=int(xml_tags['NofQuantisationLevels']['Value']) 
            meta['apply_offset'] = True
        else:
            raise(RuntimeError(f"File format (SignalFormat) is not supported.\n \
                                 Tag: {xml_tags['SignalFormat']}\n \
                                 XML file: {self.metadata_file} "))

        meta['quantisation_levels']= quantisation_levels= int(xml_tags['NofQuantisationLevels']['Value'])
        meta['encoding_format'] = encoding_format
        meta['channel_name'] = [x[0] for x in encoding_format]
        # Ensure the data format is normal. In the future other data formats might need separate hanlding.
        if self.xml_tags["TraceType"]['Value'] == 'eRS_TRACE_TYPE_NORMAL':
            pass
        elif self.xml_tags["TraceType"]['Value'] == 'eRS_TRACE_TYPE_AVERAGE':
            pass
        else:
            raise(RuntimeError(f'File format {self.xml_tags["TraceType"]["Value"]} is not supported. Have fun implementing it :)!'))

        # Output some info to the log.
        logging.info(f'RecordLength: {xml_tags["RecordLength"]}') 
        logging.info(f'TriggerOffset: {xml_tags["TriggerOffset"]}') 
        logging.info(f'SignalRecordLength: {xml_tags["SignalRecordLength"]}') 
        logging.info(f'Source: {xml_tags["Source"]}')
        
        # Save the conversion factor and offset for later processing
        meta['conversion_factor'] = {}
        meta['offset'] = {}
        if meta['apply_offset']:
            # Single and multi-channel needs seaprate handling it seems.
            if xml_tags["MultiChannelExport"]['Value'] =='eRS_ONOFF_OFF':
                # The Position is indicated in divisions. We have to convert to V
                position_div = float(xml_tags['VerticalPosition']['Value'])
                verticalScale = float(xml_tags['VerticalScale']['Value'])
                step_factor = float(xml_tags['VerticalScale']['StepFactor'])
                position = position_div * verticalScale
                # The offset is in V.
                offset = float(xml_tags['VerticalOffset']['Value'])
                meta['conversion_factor'][source_names[-1]] = 1 / float(quantisation_levels) * step_factor * verticalScale        
                meta['offset'][source_names[-1]] = offset - position
            elif xml_tags["MultiChannelExport"]['Value'] == 'eRS_ONOFF_ON':
                for sn, ch in zip(source_names, channel_names):
                    position_div = float(xml_tags['MultiChannelVerticalPosition'][ch])
                    verticalScale = float(xml_tags['MultiChannelVerticalScale'][ch])
                    step_factor = float(xml_tags['MultiChannelVerticalScale']['StepFactor'])
                    position = position_div * verticalScale
                    offset = float(xml_tags['MultiChannelVerticalOffset'][ch])
                    meta['conversion_factor'][sn] = 1 / float(quantisation_levels) * step_factor * verticalScale        
                    meta['offset'][sn] = offset - position
        
        # Info for fast segmentation
        meta['NumberOfAcquisitions']=int(xml_tags['NumberOfAcquisitions']['Value'])
        
        # Save some metadata
        meta['xStart'] = float(xml_tags['XStart']['Value'])
        meta['xStop'] = float(xml_tags['XStop']['Value'])
        meta['length_total'] = int(xml_tags['RecordLength']['Value'])*meta['NumberOfAcquisitions']
        meta['length_acquisition'] = int(xml_tags['RecordLength']['Value'])
        meta['t_sample'] = (meta['xStop']-meta['xStart'])/meta['length_acquisition']
        meta['data_format']='numpy'
        
        meta['bytes_per_sample']=bytes_per_sample
        # Bytes at the beginning of the file which do not store data. This is a magic number.
        meta['initial_offset']=8
        # We don't have this information yet, so initialize the value.
        meta['acquistions_incomplete'] = None
        return meta

    def timeToIndex(self, time):
        '''
        Returns an index corresponding to the given time in seconds. Returns None if the time is out of range.
        '''
        
        if time < self.meta['xStart'] or time >  self.meta['xStop']:
            return None
        
        time_offset=time-self.meta['xStart']
        index=time_offset/self.meta['t_sample']
        return int(index)

    def check_file(self):
        '''
        Performs some checks on the files. This includes comparing the file size of the binary file with the expected file size calculated using the 
        meta data and trying to memmap the file including sanity checks on the memmap. 
        
        :returns: size_matches, acquistions_incomplete, file_corrupted  
        
        size_matches: True if the file size matches, False if they differ.
        acquistions_incomplete: None if the file cannot be read. True if a multiple acquisitions are expected but not all are present in the file. 
        file_corrupted: True if memmapping the file and corresponding sanity checks fail. This may also indicate an unsupported file format.
        '''
        
        # Try to read the file first. This will correct the metadata for files which don't contain the complete history  
        # Reading will fail on corrupted files, therefore we ignore any failure to do so.
        file_corrupted=False
        try:
            self.getRaw()
        except:
            file_corrupted=True
        no_samples_in_file=self.meta['NumberOfAcquisitions']*self.meta['length_raw_expected']
        expected_file_size=self.meta['initial_offset']+no_samples_in_file*self.meta['bytes_per_sample']
        file_size_on_disk=os.path.getsize(self.waveform_file)
        print(f'Expected file size: {expected_file_size}, File size on diks: {file_size_on_disk}')
        
        sizes_match=expected_file_size == file_size_on_disk
        acquistions_incomplete=self.meta['acquistions_incomplete']
        
        return sizes_match, acquistions_incomplete, file_corrupted

    def check_xml(self):
        '''
        Does a simple sanity check on the xml file. A correctly written xml file must end with "</Database>".
        '''
        with open(self.metadata_file) as f:
            for line in f:
                pass
            last_line = line
        
        print(last_line)
        
        return last_line=='</Database>'

    @property
    def data_raw_RS(self):
        if self._data is None:
        # Load the data via numpy. To get a numpy array we first load as a single array.
        # Then we reshape and convert to float as we want to do some math on the data later on.
        # We seem to have 8 bytes of offset
        # Just load the data once it is really needed, not already during initializing.
            
            
            #with open(self.waveform_file,"rb") as file:
            #    self.binary_header_a=file.read(4)
            
            self.binary_header=np.fromfile(self.waveform_file, dtype='<i4', count=2, offset=0)
            #self.binary_header_b=np.fromfile(self.waveform_file, dtype='<i4', count=1, offset=4)
            
            #self.size=2**self.binary_header[0] * self.binary_header[1] 
            #print('Binary header A: {}'.format(hex(self.binary_header_a[0])))
            print('Binary header: {}, HW length: {}'.format(self.binary_header, self.meta['length_raw_expected']))
            #print('Binary multiply: {}'.format(self.binary_header[0]))
            #print('Size: {}'.format(self.size))
            print('Encoding format: {}'.format(self.meta['encoding_format']))
            
            #print('Binary header B: {}'.format(self.binary_header_b))
            #print('Binary header')
            #if not self.binary_header == b'\x04\x00\x00\x00\xe6\x0f\x00\x00':
            #    raise(RuntimeError(f'Found unexpected header. {self.metadata_file }: {self.binary_header}'))
            
            # Get the data from the file
            self.memmap = np.memmap(self.waveform_file, dtype=self.meta['encoding_format'], offset = self.meta['initial_offset'], mode='r')

            # Reshape.
            # We keep the reshaped data in memory for fast access later on.
            self._data={}
            for n in self.meta['channel_name']:
                self._data[n]=self.memmap[n]
            
            # Do some sanity check. We expect the 'Value' of "SignalHardwareRecordLength" of raw samples. 
            # Note1: At this point we expect to see the 'Value' and not the 'UserValue'.

            length_raw_expected = self.meta['length_raw_expected']*self.meta['NumberOfAcquisitions']
            length_raw_observed = len(self._data[self.meta['channel_name'][0]])
            # Doing sanity checks here is tricky.
            # 1. There seems to be a bug in the oscilloscope software. For xy data the length does not match even if the data is fine
            # 2. If NumberOfAcquisitions is larger than 1 this seems to be an upper bound and not the actual lenght of the acquisitions.
           
            self.meta['acquistions_incomplete']=False
            if length_raw_expected != length_raw_observed: # and not self.meta['ignore_lengthcheck']:
                message=f'Sanity check failed. SignalHardwareRecordLength does not match decoded data. \n length_raw_expected:{length_raw_expected}, length_raw_observed:{length_raw_observed}'
                if self.meta['NumberOfAcquisitions'] > 1:
                    if length_raw_expected > length_raw_observed and length_raw_observed%self.meta['length_raw_expected'] == 0:
                        # We ignore this case. It happens if the incomplete history is saved.
                        self.meta['length_total']=len(self.memmap)
                        self.meta['NumberOfAcquisitions']=self.meta['length_total']/self.meta['length_raw_expected']
                        self.meta['acquistions_incomplete']=True
                    else:
                        raise(RuntimeError(message))
                else:
                    raise(RuntimeError(message))
                        
            logging.info(f'Loading binary data from file {self.waveform_file} completed.')

        return self._data

    def rawToDtypeOut(self, data):
        '''
        Helper function converting the data to DTYPE_OUT and applying the conversion_factor, offset and scaling to the given data.        
        
        :param data: dict like self.data_raw
        '''
        
        if self.meta['apply_offset']:
            for key in data.keys():
                data[key]=data[key].astype(self.DTYPE_OUT)
                data[key] *= self.meta['conversion_factor'][key]
                data[key] += self.meta['offset'][key]
        
        return data

    def _gen_start_stop_idx(self, acquisition, start, stop ):
        '''
        Generate start stop indices for our memory mapped data.
        
        '''
        if not start is None and not stop is None:
            if stop <= start:
                raise(RuntimeError('stop must be larger than start'))
        
        # If the data includes multiple acquisitions, the user may restrict the data access to one acquisitions
        # Alternatively, the complete data  
        if acquisition is None:
           
            # Sanity check on start and stop
            if not stop is None:
                if stop < 0 or stop > self.samples_per_acquisition:
                    raise(RuntimeError(f'Invalid value for stop {stop}')) 
            if not start is None:
                if start < 0 or start > self.samples_per_acquisition:
                    raise(RuntimeError(f'Invalid value for stop {start}')) 
           
            # Compute start and stop indices cutting away the leading samples.  
            start_idx=self.meta['leading_samples']
            if stop is None:
                stop_idx = start_idx + self.meta['length_total']
            else:
                stop_idx = start_idx + stop  
            
            if not start is None:
                start_idx += start 
        else: 
            
            # Sanity check on start and stop
            if not stop is None:
                if stop < 0 or stop > self.meta['length_total']:
                    raise(RuntimeError(f'Invalid value for stop {stop}')) 
           
            if not start is None:
                if start < 0 or start > self.meta['length_total']:
                    raise(RuntimeError(f'Invalid value for stop {start}')) 
        
            # Sanity check on the acquisition number
            if acquisition < 0 or acquisition > self.no_acquisitions:
                raise(RuntimeError(f'Invalid  acquisition number: {acquisition}'))
            
            # Do some sanity check. I have never encountered LeadingSettlingSamples unequal 0 when  NumberOfAcquisitions was unequal to zero so I don't know how to handle this situation.
            if self.meta['leading_samples'] != 0 :
                raise(RuntimeError("Encountered LeadingSettlingSamples != 0 while trying to access a single acquisition. It's unclear how to handle this. Therefore I'm giving up")) 
            
            # Compute start and stop indices
            start_idx = acquisition*self.samples_per_acquisition
            if stop is None:
                stop_idx = start_idx + self.samples_per_acquisition 
            else:
                stop_idx = start_idx + stop
            
            if not start is None:
                start_idx += start

        return start_idx, stop_idx

    def getRaw(self, acquisition=None, start = None, stop = None, source=None):
        '''
        Filters the raw data using start stop and source.
        Returns something similar to the raw data but filtered. Time data is removed.
        
        :param acquisition: 
        :param start: Index of starting sample
        :param stop: Index of final sample
        :param source: Oscilloscope Source Name 
        '''
        
        # A bit of cryptic python code:). 
        # my_data is a dict like self.data_raw but filtered to source or meta['source_names']. 
        # This way we eliminate the time.
        if source is None:
            my_data={k:v for (k,v) in self.data_raw_RS.items() if k in self.meta['source_names']}
        else:
            my_data={k:v for (k,v) in self.data_raw_RS.items() if k in source}
        
        start_idx, stop_idx =self._gen_start_stop_idx(acquisition, start, stop)
           
        # Now apply the start stop filter
        # We will handle the time separately
        for key in my_data.keys():
            my_data[key]=my_data[key][start_idx:stop_idx]
            
        return my_data
    
    @property
    def sample_time(self):
        '''
        Duration of one sample in seconds
        '''
        return self.meta['t_sample']
        
    @property
    def all_channels(self):
        '''
        Lists all channels in the data including a time channel if available
        '''
        return self.meta['channel_name']
    
    @property
    def signal_sources(self):
        '''
        List of signal sources
        '''
        sources=self.meta['source_names']
        return sources
    
    @property
    def no_acquisitions(self):
        '''
        Number of acquisition e.g. in case a history was saved
        '''
        return self.meta['NumberOfAcquisitions']
        
    @property
    def samples_per_acquisition(self):
        '''
        Number of samples per acquisition
        '''
        return self.meta['length_acquisition']
    
    @property
    def total_no_samples(self):
        '''
        Total number of samples
        '''
        return self.meta['length_total']
    
    def getAsDf(self, acquisition=None, start = None, stop = None, source=None, time=True):
        '''
        Returns a dataframe containing the data between start and stop. By default, also timestamps will be included.
        You can set time to False to exclude timestamps.
        This is useful for plotting or other processing of large files.
        
        :param start: Index of starting sample
        :param stop: Index of final sample
        :param source: Oscilloscope Source Name 
        :param time: Optional. Defaults to True. If false, no time data is returned.
        :type time: Boolean
        '''
        # Filtering is done a separate function. We get filtered data without time.
        my_data=self.getRaw(acquisition=acquisition, start = start, stop = stop, source=source)
        
        # We don't have time information in our data yet.
        # It's time to convert and scale.
        my_data=self.rawToDtypeOut(my_data)
        
        # We now have to add the time if requested by the user.
        # If xy data was selected we get a time axis from the oscilloscope. 
        # Otherwise we have to generate it.
        if time:
            if  'X-Time' in self.meta['channel_type']:
                tsChannel=self.meta['channel_type'].index('X-Time')
                name=self.meta['channel_name'][tsChannel]
                # If we have x-y data we also have to apply the start stop filter.
                start_idx, stop_idx=self._gen_start_stop_idx(acquisition, start, stop)
                my_data[name]=self.data_raw_RS[name][start_idx:stop_idx]
            else:
                # Generate time stamps. We only generate them for the inteval between start and stop
                try:
                    timestamps=self._genTimeAxis(acquisition, start, stop)
                except:
                    print('Breakpoint')
                    raise
                my_data['Time']=timestamps
        
        # Generate a DataFrame
        try:
            df = pd.DataFrame(my_data)
        except:
            print('Breakpoint')
            raise
        
        return df

    def getLimitedStartStop(self):
        '''
        Returns start and stop value for :py:meth:`.getAsDf` or :py:meth:`.getRaw` with a limited length.
        The start stop values are centered to the 
        
        :returns: start, stop
        '''
        MAXPLOTLEN = 5E7
        
        dfLen = self.no_samples_acquisitions
        
        # If greater than the predefined length, one defines a region in the middle of the timeframe with
        # size MAXPLOTLENGTH
        if dfLen > MAXPLOTLEN:
            start = int(dfLen/2-MAXPLOTLEN/2)
            stop = int(dfLen/2+MAXPLOTLEN/2)

        # Unnecessary but doesn't make it faster anyway
        else:
            start = 0
            stop = int(dfLen)

        # Just in case: saturate
        if start < 0:
            start = 0
        if stop > dfLen:
            stop = dfLen

        
        return start, stop
    
